/**
 * 音声マネージャ
 *
 * @note 音声設定は原則的にゲームサーバの値を使用するので、
 *        アプリではlocalStorageで管理しています。
 *        もし、起動時の通信が気に入らない場合は、SharedPreferencesを基準にするようにしてください。
 *
 * @constructor
 */
var SoundManager = function () {

	this.soundEnable = 1;
	this.soundDisable = 0;
	this.states = {play: 1, other: 0};		// 再生状態
	this.playTypes = {play: 'play', loop: 'loop'};		// 再生区分
	this.soundType = {bgm: 'music', se: 'se'};		// 音声区分

	this.configKeyPrefix = 'sound_';

	/**
	 * 各種設定値のキー定義
	 * @type {string}
	 */
	this.playingListKey = 'playing_list';		// loop再生中管理用リストを保存するキー
	this.defaultMusicKey = 'defaultMusic';		// デフォルト音楽
	this.playScenesKey = "playScenes";		// 音声再生箇所
	this.excludeExecPlayScenesKey = "excludeExecPlayScenes";		// 音声再生判定除外箇所

	/**
	 * アプリ起動時に読み込む音声リスト
	 * @type {{idStr: {path: string, type: string}, ...}}
	 */
	this.loadSounds = {
		bgm_battle: {path: 'bgm/bgm_battle.m4a', type: 'music'},
		bgm_boss: {path: 'bgm/bgm_boss.m4a', type: 'music'},
		bgm_ootono_v2: {path: 'bgm/bgm_ootono_v2.m4a', type: 'music'},
		punch: {path: 'bgm/punch.mp3', type: 'se'}
	};

	/**
	 * 音声設定
	 * @type {{id: int}}
	 */
	this.defaultConfigs = {
		music: 1,
		se: 1
	};

	/**
	 * 常に流しっぱなしの音声設定
	 * @type {string}
	 */
	this.defaultMusic = '';

	/**
	 * 音声再生場面リスト
	 * ※action に "all" を指定した場合は、全てのアクションで指定音声を再生
	 *
	 * @type {{controller: {action: sound ID}}}
	 */
	this.playScenes = {	};

	/**
	 * 音声再生場面除外リスト
	 * ※action に "all" を指定した場合は、全てのアクションで判定しない
	 *
	 * @type {{controller: {action: 1}}}
	 */
	this.excludeExecPlayScenes = {};

};

/**
 * 音声管理の初期処理
 */
SoundManager.prototype.init = function() {
	PGLog.writeWithTime("SoundManager.init start");
	soundManager.initLoad();
	soundManager.initConfig();
	PGLog.writeWithTime("SoundManager.init end");
};

/**
 * 音声ファイルの初期読み込み
 */
SoundManager.prototype.initLoad = function () {
	var path = '';
	var type = '';
	for (var id in this.loadSounds) {
		path = this.loadSounds[id]['path'];
		type = this.loadSounds[id]['type'];
		PGMedia.preload(id, path, type);
	}
};

/**
 * 初期設定
 */
SoundManager.prototype.initConfig = function () {
	var self = this;
	var value = null;

	// デフォルト音楽
	this.setDefaultMusic(this.defaultMusic);

	// 音声再生箇所
	appConfig.setConfig(this.playScenesKey, JSON.stringify(this.playScenes), false);
	appConfig.setConfig(this.excludeExecPlayScenesKey, JSON.stringify(this.excludeExecPlayScenes), false);
	appConfig.setConfig('isInitSettingPlayScene', 0, false);		// 保険処理

	// 再生設定
	for (var key in self.defaultConfigs) {
		value = self.getConfig(key);
		if (value == null) {
			// なければデフォルト値を保持しておく
			self.setConfig(key, self.defaultConfigs[key]);
		}
	}

};

/**
 * 設定値を保存
 *
 * @param key
 * @param value
 */
SoundManager.prototype.setConfig = function (key, value) {
	var nowValue = this.getConfig(key);
	if (value != nowValue) {
		var configKey = this.configKeyPrefix + key;
		appConfig.setConfig(configKey, value);

		var defaultMusic = this.getDefaultMusic();
		if (value == this.soundEnable && defaultMusic != '' && key != 'se') {
			this.changeMusic(defaultMusic);		// 直近のデフォルト音楽を使う
		} else {
			this.autoStop(key);
		}
	}
};

/**
 * デフォルト音声を設定
 *
 * @param id
 */
SoundManager.prototype.setDefaultMusic = function(id) {
	appConfig.setConfig(this.defaultMusicKey, id, false);
};

/**
 * デフォルト音声を取得
 *
 * @returns {*}
 */
SoundManager.prototype.getDefaultMusic = function() {
	return appConfig.getConfig(this.defaultMusicKey, false);
}

/**
 * 設定値を取得
 *
 * @param key
 * @returns {*}
 */
SoundManager.prototype.getConfig = function (key) {
	var configKey = this.configKeyPrefix + key;
	return appConfig.getConfig(configKey);
};

/**
 * 音声再生可否
 *
 * @param id
 * @returns {boolean}
 */
SoundManager.prototype.isEnable = function (id) {
	if (id in this.loadSounds) {
		var type = this.loadSounds[id]['type'];
		var value = this.getConfig(type);
		if (value == this.soundEnable) {
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
};

/**
 * 指定IDの音声を再生
 *
 * @param id
 */
SoundManager.prototype.play = function (id) {
	if (this.isEnable(id)) {
		PGMedia.play(id);
	}
};

/**
 * 指定IDの音声をループ再生
 *
 * @param id
 */
SoundManager.prototype.loop = function (id) {
	if (this.isEnable(id)) {
		PGMedia.loop(id);
		this.setPlaying(id);
	}
};

/**
 * レジューム時用音声再生
 *
 * @param id
 */
SoundManager.prototype.playForResume = function (id) {
	if (this.isEnable(id)) {
		PGMedia.playForResume(id);
		this.setPlaying(id);
	}
};

/**
 * 指定IDの音声を停止
 *
 * @param id
 */
SoundManager.prototype.stop = function (id) {
	PGMedia.stop(id);
	this.removePlaying(id);
};

/**
 * 指定IDの音声を一時停止
 *
 * @param id
 */
SoundManager.prototype.pause = function (id) {
	PGMedia.pause(id);
};

/**
 * 指定IDの音声再生中判別
 *
 * @param id
 * @returns {boolean}
 */
SoundManager.prototype.isPlaying = function (id) {
	var playingList = PGLocalStorage.getVal(this.playingListKey);

	if (!playingList) return false;

	playingList = JSON.parse(playingList);
	if (id in playingList) {
		if (playingList[id]['state'] != this.states['play']) return false;
	} else {
		return false;
	}

	return true;
};

/**
 * デフォルト音声の変更
 *
 * @param id
 */
SoundManager.prototype.changeMusic = function (id) {
	if (typeof id === 'undefined') {
		for(var soundType in this.defaultConfigs) {
			this.autoStop(soundType);
		}
		this.setDefaultMusic('');
	} else {
		if (!this.isPlaying(id)) {
			// 再生中のものを全部止めてから再生させる
			for(var soundType in this.defaultConfigs) {
				this.autoStop(soundType);
			}
			this.loop(id);
			this.setDefaultMusic(id);
		}
	}
};

/**
 * 指定IDの音声を再生中音声リストに追加
 *
 * @param id
 * @param playType
 * @returns {boolean}
 */
SoundManager.prototype.setPlaying = function (id, playType) {
	if (typeof playType === 'undefined') playType = this.playTypes['loop'];
	if (id in this.loadSounds){ } else { return false; }
	var soundType = this.loadSounds[id]['type'];
	var playingList = PGLocalStorage.getVal(this.playingListKey);

	if (!playingList) {
		playingList = {};
	} else {
		playingList = JSON.parse(playingList);
	}
	playingList[id] = {state: this.states['play'], type: playType, soundType: soundType};

	PGLocalStorage.setVal(this.playingListKey, JSON.stringify(playingList));
	return true;
};

/**
 * 指定IDの音声を再生中音声リストから除去
 *
 * @param id
 * @returns {boolean}
 */
SoundManager.prototype.removePlaying = function (id) {
	var playingList = PGLocalStorage.getVal(this.playingListKey);

	if (!playingList) return true;
	if (playingList) {
		playingList = JSON.parse(playingList);
		if (id in playingList) {
			delete playingList[id];
			PGLocalStorage.setVal(this.playingListKey, JSON.stringify(playingList));
		}
	}

	return true;
};

/**
 * 全一時停止
 *
 * @returns {boolean}
 */
SoundManager.prototype.autoPause = function () {
	var playingList = PGLocalStorage.getVal(this.playingListKey);

	if (!playingList) return true;
	playingList = JSON.parse(playingList);
	for (var id in playingList) {
		if (playingList[id]['type'] == this.playTypes['loop']) {
			this.pause(id);
		} else {
			this.stop(id);
		}
	}
};

/**
 * 全停止
 * @param soundType
 * @returns {boolean}
 */
SoundManager.prototype.autoStop = function (soundType) {
	var playingList = PGLocalStorage.getVal(this.playingListKey);

	if (!playingList) return true;
	playingList = JSON.parse(playingList);
	for (var id in playingList) {
		if (soundType == playingList[id]['soundType']) this.stop(id);
	}
};

/**
 * 復帰
 *
 * @returns {boolean}
 */
SoundManager.prototype.autoResume = function () {
	var playingList = PGLocalStorage.getVal(this.playingListKey);

	if (!playingList) return true;
	playingList = JSON.parse(playingList);
	for (var id in playingList) {
		this.playForResume(id);
	}
};

/**
 * 指定音声再生シーンで音声IDを取得
 *
 * @param controller
 * @param action
 * @param paramStr
 * @returns {*}
 */
SoundManager.prototype.getSoundIdByPlayScene = function(controller, action, paramStr) {
	if (Object.keys(this.playScenes).length < 1) {
		this.playScenes = JSON.parse(appConfig.getConfig(this.playScenesKey, false));
	}

	if (typeof paramStr == 'undefined') paramStr = "";

	if (controller in this.playScenes) {
		if (action in this.playScenes[controller]) {
			return this.playScenes[controller][action];
		} else if (paramStr != "" && action + paramStr in this.playScenes[controller]) {		// チュートリアル対応
			return this.playScenes[controller][action + paramStr];
		} else if ("all" in this.playScenes[controller]) {
			return this.playScenes[controller]["all"];
		}
	}
	return null;
};

/**
 * 音声再生シーンを追加
 *
 * @param controller
 * @param action
 * @param id
 */
SoundManager.prototype.setPlayScene = function(controller, action, id) {
	this.playScenes = JSON.parse(appConfig.getConfig(this.playScenesKey, false));
	if (!(controller in this.playScenes)) {
		this.playScenes[controller] = {};
	}
	this.playScenes[controller][action] = id;
	appConfig.setConfig(this.playScenesKey, JSON.stringify(this.playScenes), false);
};

/**
 * 音声再生シーンを一括追加
 * リスト例：{{controller name: {action name: sound name, ...}, ...}}
 *
 * @param scenes
 * @returns {boolean}
 */
SoundManager.prototype.setPlayScenes = function(scenes) {
	if (typeof scenes === 'undefined') return true;

	var actions = {};
	for (var controller in scenes) {
		//PGLog.write("addPlayScenes: controller " + controller);
		actions = scenes[controller];
		for (var action in actions) {
			//PGLog.write("addPlayScenes: action " + action + "/ id " + actions[action]);
			this.setPlayScene(controller, action, actions[action]);
		}
	}
};

/**
 * 指定コントローラ（＆アクション）で音声再生シーンを除去
 *
 * @param controller
 * @param action
 */
SoundManager.prototype.deletePlayScene = function(controller, action) {
	this.playScenes = JSON.parse(appConfig.getConfig(this.playScenesKey, false));
	if (typeof action === 'undefined') {
		delete this.playScenes[controller];
	} else {
		delete this.playScenes[controller][action];
	}
	appConfig.setConfig(this.playScenesKey, JSON.stringify(this.playScenes), false);
};

/**
 * 音声再生除外シーン?
 *
 * @param controller
 * @param action
 * @returns {boolean}
 */
SoundManager.prototype.isExcludePlayScene = function(controller, action) {
	if (Object.keys(this.excludeExecPlayScenes).length < 1) {
		this.excludeExecPlayScenes = JSON.parse(appConfig.getConfig(this.excludeExecPlayScenesKey, false));
	}

	return controller in this.excludeExecPlayScenes && (action in this.excludeExecPlayScenes[controller] || "all" in this.excludeExecPlayScenes[controller]);
};

/**
 * 音声再生除外シーンを追加
 *
 * @param controller
 * @param action
 */
SoundManager.prototype.setExcludePlayScene = function(controller, action) {
	this.excludeExecPlayScenes = JSON.parse(appConfig.getConfig(this.excludeExecPlayScenesKey, false));
	if (!(controller in this.excludeExecPlayScenes)) {
		this.excludeExecPlayScenes[controller] = {};
	}
	this.excludeExecPlayScenes[controller][action] = 1;
	appConfig.setConfig(this.excludeExecPlayScenesKey, JSON.stringify(this.excludeExecPlayScenes), false);
};

/**
 * 音声再生除外シーンを一括追加
 * リスト例：{{controller name: {action name: 1, ...}, ...}}
 *
 * @param scenes
 * @returns {boolean}
 */
SoundManager.prototype.setExcludePlayScenes = function(scenes) {
	if (typeof scenes === 'undefined') return true;

	var actions = {};
	for (var controller in scenes) {
		//PGLog.write("addExcludePlayScenes: controller " + controller);
		actions = scenes[controller];
		for (var action in actions) {
			//PGLog.write("addExcludePlayScenes: action " + action);
			this.setExcludePlayScene(controller, action);
		}
	}
};

/**
 * 指定コントローラ（＆アクション）で音声再生除外シーンを除去
 *
 * @param controller
 * @param action
 */
SoundManager.prototype.deleteExcludePlayScene = function(controller, action) {
	this.excludeExecPlayScenes = JSON.parse(appConfig.getConfig(this.excludeExecPlayScenesKey, false));
	if (typeof action === 'undefined') {
		delete this.excludeExecPlayScenes[controller];
	} else {
		delete this.excludeExecPlayScenes[controller][action];
	}
	appConfig.setConfig(this.excludeExecPlayScenesKey, JSON.stringify(this.excludeExecPlayScenes), false);
};

/**
 * 指定URLで音声を再生
 *
 * @param url
 */
SoundManager.prototype.execPlayScene = function(url) {
	var elm = document.createElement('a');
	elm.href = url;

	// controller/action を取得
	var pathList = elm.pathname.split('/');
	var controller = "homes";
	var action = "index";
	var paramStr = "";
	var i = 0;
	if (pathList.length > 1) {
		var controllerIndex = 1;
		controller = pathList[controllerIndex];
		if (controller.search(/\.php$/) != -1) {
			controllerIndex++;
			controller = pathList[controllerIndex];
		}

		// controller名をcake FW仕様に沿うように変換する
		if (controller.search(/_/) != -1) {
			// スネークケース
			var words = controller.split('_');
			controller = "";
			for(i = 0; i < words.length; i++) {
				controller += words[i].toUpperCase();
			}
		} else {
			// キャメルケース＆パスカルケース
			controller = controller.charAt(0).toUpperCase() + controller.slice(1);
		}

		var actionIndex = controllerIndex + 1;
		if (actionIndex in pathList) action = pathList[actionIndex];

		// チュートリアル対応
		var paramIndex = actionIndex + 1;
		if (paramIndex in pathList) {
			for(i = paramIndex; i < pathList.length; i++) {
				paramStr += "/" + pathList[i];
			}
		}
	}

	// 拡張子が画像、css、jsなら終了
	var check = (pathList.length > 0)? pathList[pathList.length - 1]: "";
	if (check != "" && check.search(/\.(?:png|jpg|jpeg|gif|css|js)$/) != -1) {
		return false;
	}

	// 実行除外対象なら終了
	if (this.isExcludePlayScene(controller, action)) {
		return false;
	}

	// あったら再生、なければデフォルトを再生
	var soundId = this.getSoundIdByPlayScene(controller, action, paramStr);
	if (soundId != null && soundId != "") {
		if (!this.isPlaying(soundId)) this.loop(soundId);		// 再生中ならなにもしない
		return true;
	}

	var defaultMusic = this.getDefaultMusic();
	if (defaultMusic != '') {
		this.loop(defaultMusic);
	} else {
		this.autoStop(this.soundType.bgm);
	}

	return true;
};

/**
 * 音声再生シーンの初期設定
 *
 * @returns {boolean}
 */
SoundManager.prototype.initSettingPlayScene = function() {
	var isInitSettingPlayScene = appConfig.getConfig('isInitSettingPlayScene', false);
	if (isInitSettingPlayScene != 1) {
		var post_data = {
			'is_yapp':1
		};

		var axi_server_url = PGLocalStorage.getVal('axi_server_url');
		var request_url = axi_server_url + "/YappHomes/get_init_play_scenes";
		PGLog.write("start init setting play scene: " + request_url);

		// Bridgeのユーザー情報取得
		ServerInquire.postData(request_url, post_data,
			function (data){
				if(data.response === 'success') {
					var scenes = data.scenes;
					soundManager.setPlayScenes(scenes);
					var excludes = data.excludes;
					soundManager.setExcludePlayScenes(excludes);
				}
			},
			function (status, error){
				PGLog.write("POST DATA ERROR.Status:" + status + "/Err:" + error + "/Url:" + request_url);
			}
		);
		appConfig.setConfig('isInitSettingPlayScene', 1, false);
	}
	return true;
};

if (soundManager == null) {
	var soundManager = new SoundManager();
}
